﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Newtonsoft.Json;

namespace MontoyaTech.Rest.Net
{
    /// <summary>
    /// A set of extensions to help with HttpListenerResponses.
    /// </summary>
    public static class HttpListenerResponseExtensions
    {
        /// <summary>
        /// Sets the response content type to text and writes the given text to it.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="text"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithText(this HttpListenerResponse response, string text)
        {
            response.ContentType = "text/plain; charset=utf-16";

            var bytes = Encoding.Unicode.GetBytes(text);
            response.OutputStream.Write(bytes, 0, bytes.Length);

            return response;
        }

        /// <summary>
        /// Sets the response content type to json and serializes the object as json and writes it.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="obj"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithJson(this HttpListenerResponse response, object obj)
        {
            response.ContentType = "application/json; charset=utf-16";

            var bytes = Encoding.Unicode.GetBytes(JsonConvert.SerializeObject(obj));
            response.OutputStream.Write(bytes, 0, bytes.Length);

            return response;
        }

        /// <summary>
        /// Sets the response content type to a file and writes the given file content to the response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="filePath"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithFile(this HttpListenerResponse response, string filePath)
        {
            response.ContentType = "application/octet-stream";
            response.Headers.Add("Content-Deposition", $@"attachment; filename=""{Path.GetFileName(filePath)}""");
            response.SendChunked = true;

            using (var fileStream = new FileStream(filePath, FileMode.Open, FileAccess.Read))
                using (var responseStream = response.OutputStream)
                    fileStream.CopyTo(responseStream);

            return response;
        }

        /// <summary>
        /// Sets the status code for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="status"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithStatus(this HttpListenerResponse response, HttpStatusCode status)
        {
            try
            {
                response.StatusCode = (int)status;
                return response;
            }
            catch
            {
                return response;
            }
        }

        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name">The name of the cookie</param>
        /// <param name="value">The value of the cookie</param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, string name, string value)
        {
            response.SetCookie(new Cookie(name, value));

            return response;
        }

        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <param name="expires"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, string name, string value, DateTime expires)
        {
            response.SetCookie(new Cookie(name, value) { Expires = expires });

            return response;
        }
    
        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <param name="httpOnly"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, string name, string value, bool httpOnly)
        {
            response.SetCookie(new Cookie(name, value) { HttpOnly = httpOnly });

            return response;
        }

        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <param name="httpOnly"></param>
        /// <param name="secure"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, string name, string value, bool httpOnly, bool secure)
        {
            response.SetCookie(new Cookie(name, value) { HttpOnly = httpOnly, Secure = secure });

            return response;
        }

        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <param name="httpOnly"></param>
        /// <param name="secure"></param>
        /// <param name="expires"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, string name, string value, bool httpOnly, bool secure, DateTime expires)
        {
            response.SetCookie(new Cookie(name, value) { HttpOnly = httpOnly, Secure = secure, Expires = expires });

            return response;
        }

        /// <summary>
        /// Sets a cookie for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="cookie"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithCookie(this HttpListenerResponse response, Cookie cookie)
        {
            response.SetCookie(cookie);

            return response;
        }

        /// <summary>
        /// Sets a header for a given response.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="name"></param>
        /// <param name="value"></param>
        /// <returns>This response.</returns>
        public static HttpListenerResponse WithHeader(this HttpListenerResponse response, string name, string value)
        {
            response.AddHeader(name, value);

            return response;
        }
    }
}
